#if 0
g++ -O0 -no-pie -ggdb3 -lm -lgsl -lgslcblas -otestsin testsin.cpp
exit 0
#endif
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <gsl/gsl_fft_real.h>
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_vector_double.h>

#define M_PI           3.14159265358979323846

struct WavHeader {
  unsigned char           RIFF[4];        // header type. must be RIFF
  unsigned char           RIFF_l[4];      // file size in little endian
  unsigned char           WAVE[4];        // file type. must be WAVE
  unsigned char           fmt_[4];        // format chunk id. Must be fmt_
  unsigned char           fmt_l[4];       // size of the of the format chunk in bytes, little endian
  unsigned char           ext[2];         // no idea what this is. Must be 0x01,0x00
  unsigned char           numch[2];       // number of channels. We only write one
  unsigned char           rate[4];        // sample rate in Hz.
  unsigned char           bytespsec[4];   // bytes per second
  unsigned char           bytespersam[2]; // bytes per sample.
  unsigned char           bitspersam[2];  // bits per sample. Here always eight.
  unsigned char           data[4];        // data chunk id. Must be data.
  unsigned char           data_l[4];      // length of the data chunk in bytes
};

/*
** Read the wav header, return the number of data bytes
*/
int readwavheader(FILE *file,struct WavHeader *wav,int *freq,int *channels)
{
  if (fread(wav,sizeof(*wav),1,file) != 1) {
    fprintf(stderr,"could not read the WAV header\n");
    return 0;
  }

  if (wav->RIFF[0] != 'R' || wav->RIFF[1] != 'I' || wav->RIFF[2] != 'F' || wav->RIFF[3] != 'F' ||
      wav->WAVE[0] != 'W' || wav->WAVE[1] != 'A' || wav->WAVE[2] != 'V' || wav->WAVE[3] != 'E' ||
      wav->fmt_[0] != 'f' || wav->fmt_[1] != 'm' || wav->fmt_[2] != 't' || wav->fmt_[3] != ' '  ||
      wav->data[0] != 'd' || wav->data[1] != 'a' || wav->data[2] != 't' || wav->data[3] != 'a') {
    fprintf(stderr,"input file is not a WAV file\n");
    return 0;
  }
  
  if (wav->ext[0] != 1 || wav->ext[1] != 0) {
    fprintf(stderr,"WAV content is not PCM compressed\n");
    return 0;
  }

  *channels = wav->numch[0] | (wav->numch[1] << 8);
  *freq     = wav->rate[0]  | (wav->rate[1] << 8) | (wav->rate[2] << 16) | (wav->rate[3] << 24);

  if (wav->bitspersam[0] != 16 || wav->bitspersam[1] != 0) {
    fprintf(stderr,"WAV precision is not 16 bits per sample\n");
    return 0;
  }
  
  return wav->data_l[0] | (wav->data_l[1] << 8) | (wav->data_l[2] << 16) | (wav->data_l[3] << 24);
}

struct parms {
  double *buffer;
  int     min,max;
};

double func(const gsl_vector * x, void * params)
{
  double phase    = gsl_vector_get(x,0);
  double period   = gsl_vector_get(x,1);
  double scale    = gsl_vector_get(x,2);
  double offset   = gsl_vector_get(x,3);
  double sum      = 0.0;
  struct parms *p = (struct parms *)params;
  int i,j;

  for(i = p->min,j = 0;i < p->max; i++,j++) {
    double v = scale * sin(2.0 * M_PI * j / period - phase) + offset;
    double w = p->buffer[i];
    sum     += (v - w) * (v - w);
  }
#if 0
  if (fabs(offset) > 40.0) {
    double delta = fabs(offset) - 40;
    sum += delta * delta * 1e6;
  }
  if (fabs(phase) > M_PI) {
    double delta = fabs(phase) - M_PI;
    sum += delta * delta * 1e6;
  }
#endif

  return sum;
}
 
int readsamples(FILE *file,int freq)
{
  struct WavHeader wav;
  int i,j,k,lo,hi,size,samplefreq,channels;
  double *buffer;
  double *residual;
  double *ref;
  double max   = 0.0;
  double signal = 0.0;
  double noise  = 0.0;
  double scale1,scale2,scale,phase,phase2;
  double dperiod;
  double offset = 0.0;
  int maxpos   = 0;
  int startpos = 0;
  int endpos   = 0;
  int fftsize,lastfftsize = 0;
  int crossings = 0;
  int period;
  int peak;
  int count;
  gsl_fft_real_wavetable  *realwav;
  gsl_fft_real_workspace  *workspace;
  gsl_multimin_fminimizer *minimizer;
  const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex2;
  gsl_vector *startpoint, *stepsize;
  gsl_multimin_function minex;

  size = readwavheader(file,&wav,&samplefreq,&channels);
  if (size <= 0)
    return 0;

  if (channels != 1 && channels != 2) {
    fprintf(stderr,"Number of channels must be 1 or 2\n");
    return 0;
  }

  buffer = new double[size >>= ((channels == 1)?1:2)];
  for(i = 0;i < size;i++) {
    int lo,hi;
    short v;
    double d;
    lo = getc(file);
    hi = getc(file);
    v  = lo | (hi << 8);
    d  = v;
    if (channels == 2) {
      lo = getc(file);
      hi = getc(file);
      v  = lo | (hi << 8);
      d  = (d + double(v)) * 0.5;
    }
    buffer[i] = d;
  }

  period = samplefreq / freq;
  /*
  ** find the first maximum 
  */
  for(i = 0;i < size - 2;i++) {
    double v = buffer[i];
    if (v > 2048 && v > max) {
      max = v;
      if (buffer[i+1] < max && buffer[i+2] < max) {
	maxpos = i;
      }
    }
    if (maxpos && buffer[i] >= 0 && buffer[i+1] < 0) {
      startpos = i;
      break;
    }
  }

  printf("Period: %d\n",period);
  endpos   = startpos + 14 * 44100; /* 15 seconds */
  if (endpos + period >= size)
    endpos = size - period;
  maxpos   = startpos + (period >> 2);

  /*
  ** find a better end position
  */
  while(endpos > startpos && buffer[endpos] < 0) {
    endpos--;
  }
  while(endpos < size && buffer[endpos] >= 0) {
    endpos++;
  }
  endpos--;
  
  if (endpos <= startpos + 2) {
    fprintf(stderr,"Cannot find reasonably sized samples\n");
    return 0;
  }


  fftsize   = endpos - startpos;
  residual  = new double[(fftsize + 1) & -2];

  for(i = 0;i < fftsize-2;i++) {
    if (buffer[i+startpos] >= 0 && buffer[i+startpos-1] >= 0 &&
	buffer[i+startpos + 1] < 0 && buffer[i+startpos + 2] < 0)
      crossings++;
  }
  printf("Period: %g\n",dperiod = fftsize * 1.0 / crossings);

#if 0
  FILE *out = fopen("out.raw","wb");
  minimizer  = gsl_multimin_fminimizer_alloc(T,4);
  startpoint = gsl_vector_alloc(4);
  stepsize   = gsl_vector_alloc(4);
  
  for(i = period * 150;i < fftsize - 20 * period;i += period * 20) {
    double delta,n=0.0,size;
    int iter = 0;
    struct parms p;
    scale1  = buffer[i + startpos] - buffer[i + startpos + (period >> 1)];
    scale2  = buffer[i + maxpos  ] - buffer[i + maxpos   + (period >> 1)];
    scale1 /= 2;
    scale2 /= -2;
    scale   = sqrt(scale1 * scale1 + scale2 * scale2);
    phase   = atan2(scale1,scale2);

    if (fabs(offset) > 5.0)
      offset = 0.0;
    
    gsl_vector_set(startpoint,0,phase);
    gsl_vector_set(startpoint,1,dperiod);
    gsl_vector_set(startpoint,2,-scale);
    gsl_vector_set(startpoint,3,offset);
    
    gsl_vector_set(stepsize,0,0.000001);
    gsl_vector_set(stepsize,1,0.000001);
    gsl_vector_set(stepsize,2,0.001);
    gsl_vector_set(stepsize,3,0.001);

    p.buffer = buffer;
    p.min    = i + startpos;
    p.max    = i + startpos + period * 20;

    minex.n      = 4;
    minex.f      = &func;
    minex.params = &p;

    gsl_multimin_fminimizer_set(minimizer,&minex,startpoint,stepsize);
    do {
      int status = gsl_multimin_fminimizer_iterate(minimizer);
      
      iter++;
      if (status) {
	fprintf(stderr,"iteration aborted\n");
	return 0;
      }

      size   = gsl_multimin_fminimizer_size (minimizer);
      status = gsl_multimin_test_size (size, 1e-8);

      if (status == GSL_SUCCESS)
	break;
    } while(iter < (1 << 17));

    noise  += gsl_multimin_fminimizer_minimum(minimizer);
    {
      gsl_vector *x   = gsl_multimin_fminimizer_x(minimizer);
      phase           = gsl_vector_get(x,0);
      dperiod         = gsl_vector_get(x,1);
      scale           = gsl_vector_get(x,2);
      offset          = gsl_vector_get(x,3);
      for(int t = p.min,j = 0;t < p.max; t++,j++) {
	double v      = scale * sin(2.0 * M_PI * j / dperiod - phase) + offset;
	double w      = p.buffer[t];
	residual[t - startpos] = v - w;
	signal       += v * v;
	{
	  int diff    = 64 * (v - w);
	  int lo      = diff >> 8;
	  int hi      = diff & 0xff;
	  fputc(lo,out);
	  fputc(hi,out);
	}
      }
    }
    printf("%d\t%g\t%g\t%g\t%g\t%g\n",i / period,scale,phase,dperiod,scale,offset);
  }
  fclose(out);
  
  printf("SNR: %gdB\n",-10.0*log(noise / signal) / log(10.0));
#endif
  
#if 1
  FILE *out = fopen("fft.plot","w");
  int windowsize = 100;
  signal = noise = 0.0;
  for(i = period * 150 + startpos;i < endpos - windowsize * period;i += period * 25) {
    int start,end;
    start  = i;
    while(buffer[start] < 0.0)
      start--;
    while(buffer[start + 1] >= 0.0)
      start++;
    end    = start + period * windowsize;
    while(buffer[end] < 0.0)
      end--;
    while(buffer[end + 1] >= 0.0)
      end++;
    fftsize   = end - start;
    printf("%d\n",fftsize);
    double winscale = 2.0 * M_PI / fftsize;
    double winshift = 0.5 * fftsize;
    for(j = 0;j < fftsize;j++) {
      double m    = 0.54 + 0.46 * cos(winscale * (j - winshift));
      residual[j] = m * buffer[j + startpos];
    }
    if (fftsize != lastfftsize) {
      if (lastfftsize) {
	gsl_fft_real_wavetable_free(realwav);
	gsl_fft_real_workspace_free(workspace);
      }
      lastfftsize = fftsize;
      realwav     = gsl_fft_real_wavetable_alloc(fftsize);
      workspace   = gsl_fft_real_workspace_alloc(fftsize);
    }
    gsl_fft_real_transform(residual,1,fftsize,realwav,
			   workspace);
    {
      double peak = fftsize * 2.0 / period;
      double maxv = peak / 16;
      double tsig = 0.0;
      double tnos = 0.0;
      int nbins   = 0;
      int sbins   = 0;
      for(j = 0;j < fftsize;j+=2) {
	double real = residual[j];
	double img  = residual[j+1];
	double abs  = real * real + img * img;
	double delt = fabs(peak - j);
	if (j > 0 && delt >= maxv) {
	  tnos     += abs;
	  nbins++;
	} else if (j > 0) {
	  tsig     += abs;
	  sbins++;
	}
	fprintf(out,"%d\t%g\n",j >> 1 ,abs);
      }
      signal += tsig;
      noise  += tnos;
    }
  }
  fclose(out);

  printf("SNR: %gdB\n",-10.0*log(noise / signal) / log(10.0));
#endif
  
  return 1;
}

  
int main(int argc,char **argv)
{
  int freq;
  
  if (argc != 2) {
    fprintf(stderr,"Usage: %s freq\n",argv[0]);
    return 5;
  }

  freq = atoi(argv[1]);
  return readsamples(stdin,freq);
}

    
  
